# -*- coding: utf-8 -*-
# Copyright (C) 2018 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#         Remi Duraffort <remi.duraffort@linaro.org>
#
# This file is part of LAVA.
#
# LAVA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# LAVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with LAVA.  If not, see <http://www.gnu.org/licenses/>.

from django import forms
from django.core.exceptions import ValidationError
from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.db import transaction
from django.conf import settings

from lava_scheduler_app.models import (
    Architecture,
    Alias,
    BitWidth,
    Core,
    DefaultDeviceOwner,
    Device,
    DeviceType,
    JobFailureTag,
    NotificationRecipient,
    ProcessorFamily,
    Tag,
    TestJob,
    User,
    Worker,
)
from linaro_django_xmlrpc.models import AuthToken

# django admin API itself isn't pylint clean, so some settings must be suppressed.
# pylint: disable=no-self-use,function-redefined


class AliasAdmin(admin.ModelAdmin):
    def get_readonly_fields(self, _, obj=None):
        if obj:  # editing an existing object
            return self.readonly_fields + ("name",)
        return self.readonly_fields


class ArchitectureAdmin(admin.ModelAdmin):
    def get_readonly_fields(self, _, obj=None):
        if obj:  # editing an existing object
            return self.readonly_fields + ("name",)
        return self.readonly_fields


class BitWidthAdmin(admin.ModelAdmin):
    def get_readonly_fields(self, _, obj=None):
        if obj:  # editing an existing object
            return self.readonly_fields + ("width",)
        return self.readonly_fields


class CoreAdmin(admin.ModelAdmin):
    def get_readonly_fields(self, _, obj=None):
        if obj:  # editing an existing object
            return self.readonly_fields + ("name",)
        return self.readonly_fields


class DefaultOwnerInline(admin.StackedInline):
    """
    Exposes the default owner override class
    in the Django admin interface
    """

    model = DefaultDeviceOwner
    can_delete = False


def expire_user_action(
    modeladmin, request, queryset
):  # pylint: disable=unused-argument
    for user in queryset.filter(is_active=True):
        AuthToken.objects.filter(user=user).delete()
        user.is_staff = False
        user.is_superuser = False
        user.is_active = False
        for group in user.groups.all():
            group.user_set.remove(user)
        for permission in user.user_permissions.all():
            user.user_permissions.remove(permission)
        user.save()


expire_user_action.short_description = "Expire user account"


class UserAdmin(UserAdmin):
    """
    Defines the override class for DefaultOwnerInline
    """

    inlines = (DefaultOwnerInline,)
    actions = [expire_user_action]

    def has_delete_permission(self, request, obj=None):
        return settings.ALLOW_ADMIN_DELETE


#  Setup the override in the django admin interface at startup.
admin.site.unregister(User)
admin.site.register(User, UserAdmin)


def cancel_action(modeladmin, request, queryset):  # pylint: disable=unused-argument
    with transaction.atomic():
        for testjob in queryset.select_for_update():
            if testjob.can_cancel(request.user):
                if testjob.is_multinode:
                    for job in testjob.sub_jobs_list:
                        job.go_state_canceling()
                        job.save()
                else:
                    testjob.go_state_canceling()
                    testjob.save()


cancel_action.short_description = "cancel selected jobs"


def fail_action(modeladmin, request, queryset):
    if request.user.is_superuser:
        with transaction.atomic():
            for testjob in queryset.select_for_update().filter(
                state=TestJob.STATE_CANCELING
            ):
                testjob.go_state_finished(TestJob.HEALTH_INCOMPLETE)
                testjob.save()


fail_action.short_description = "fail selected jobs"


class ActiveDevicesFilter(admin.SimpleListFilter):
    title = "Active devices"
    parameter_name = "state"

    def lookups(self, request, model_admin):
        return (("NoRetired", "Exclude retired"), ("CurrentJob", "With a current Job"))

    def queryset(self, request, queryset):
        if self.value() == "NoRetired":
            return queryset.exclude(health=Device.HEALTH_RETIRED).order_by("hostname")
        if self.value() == "CurrentJob":
            return queryset.filter(
                state__in=[Device.STATE_RESERVED, Device.STATE_RUNNING]
            ).order_by("hostname")


class ActualDeviceFilter(admin.SimpleListFilter):
    title = "Actual Device (except retired)"
    parameter_name = "actual_device"

    def lookups(self, request, model_admin):
        list_of_types = []
        queryset = Device.objects.exclude(health=Device.HEALTH_RETIRED).order_by(
            "hostname"
        )
        for dev_type in queryset:
            list_of_types.append((str(dev_type.hostname), dev_type.hostname))
        return sorted(list_of_types, key=lambda tp: tp[1])

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(actual_device__hostname=self.value())
        return queryset.order_by("actual_device__hostname")


class DeviceTypeFilter(admin.SimpleListFilter):
    title = "Device Type"
    parameter_name = "device_type"

    def lookups(self, request, model_admin):
        list_of_types = []
        queryset = DeviceType.objects.all()
        for dev_type in queryset:
            list_of_types.append((str(dev_type.name), dev_type.name))
        return sorted(list_of_types, key=lambda tp: tp[1])

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(device_type__name=self.value())
        return queryset.order_by("device_type__name")


class RequestedDeviceTypeFilter(admin.SimpleListFilter):
    title = "Requested Device Type"
    parameter_name = "requested_device_type"

    def lookups(self, request, model_admin):
        list_of_types = []
        queryset = DeviceType.objects.order_by("name")
        for dev_type in queryset:
            list_of_types.append((str(dev_type.name), dev_type.name))
        return sorted(list_of_types, key=lambda tp: tp[1])

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(requested_device_type__name=self.value())
        return queryset.order_by("requested_device_type__name")


def _update_devices_health(request, queryset, health):
    with transaction.atomic():
        for device in queryset.select_for_update():
            old_health_display = device.get_health_display()
            device.health = health
            device.save()
            device.log_admin_entry(
                request.user,
                "%s → %s" % (old_health_display, device.get_health_display()),
            )


def device_health_good(modeladmin, request, queryset):
    _update_devices_health(request, queryset, Device.HEALTH_GOOD)


def device_health_unknown(modeladmin, request, queryset):
    _update_devices_health(request, queryset, Device.HEALTH_UNKNOWN)


def device_health_maintenance(modeladmin, request, queryset):
    _update_devices_health(request, queryset, Device.HEALTH_MAINTENANCE)


def device_health_retired(modeladmin, request, queryset):
    _update_devices_health(request, queryset, Device.HEALTH_RETIRED)


device_health_good.short_description = "Update health of selected devices to Good"
device_health_unknown.short_description = "Update health of selected devices to Unknown"
device_health_maintenance.short_description = (
    "Update health of selected devices to Maintenance"
)
device_health_retired.short_description = "Update health of selected devices to Retired"


class DeviceAdmin(admin.ModelAdmin):
    list_filter = (
        DeviceTypeFilter,
        "state",
        ActiveDevicesFilter,
        "health",
        "worker_host",
    )
    raw_id_fields = ["last_health_report_job"]

    def has_health_check(self, obj):
        return bool(obj.get_health_check())

    has_health_check.boolean = True
    has_health_check.short_description = "HC"

    def health_check_enabled(self, obj):
        return not obj.device_type.disable_health_check

    health_check_enabled.boolean = True
    health_check_enabled.short_description = "HC enabled"

    def get_readonly_fields(self, _, obj=None):
        if obj:  # editing an existing object
            return self.readonly_fields + ("hostname",)
        return self.readonly_fields

    def valid_device(self, obj):
        return bool(obj.is_valid())

    valid_device.boolean = True
    valid_device.short_description = "Config"

    def device_dictionary_jinja(self, obj):
        return obj.load_configuration(output_format="raw")

    def has_delete_permission(self, request, obj=None):
        return settings.ALLOW_ADMIN_DELETE

    fieldsets = (
        (
            "Properties",
            {"fields": ("hostname", "device_type", "worker_host", "device_version")},
        ),
        (
            "Device owner",
            {
                "fields": (
                    ("user", "group"),
                    ("physical_owner", "physical_group"),
                    "is_public",
                )
            },
        ),
        (
            "Status",
            {
                "fields": (
                    ("state", "health"),
                    ("last_health_report_job", "current_job"),
                )
            },
        ),
        (
            "Advanced properties",
            {
                "fields": ("description", "tags", ("device_dictionary_jinja")),
                "classes": ("collapse",),
            },
        ),
    )
    readonly_fields = ("device_dictionary_jinja", "state", "current_job")
    list_display = (
        "hostname",
        "device_type",
        "current_job",
        "worker_host",
        "state",
        "health",
        "has_health_check",
        "health_check_enabled",
        "is_public",
        "valid_device",
    )
    search_fields = ("hostname", "device_type__name")
    ordering = ["hostname"]
    actions = [
        device_health_good,
        device_health_unknown,
        device_health_maintenance,
        device_health_retired,
    ]


class VisibilityForm(forms.ModelForm):
    def clean_viewing_groups(self):
        viewing_groups = self.cleaned_data["viewing_groups"]
        visibility = self.cleaned_data["visibility"]
        if len(viewing_groups) != 1 and visibility == TestJob.VISIBLE_GROUP:
            raise ValidationError(
                "Group visibility must have exactly one viewing group."
            )
        elif viewing_groups and visibility == TestJob.VISIBLE_PERSONAL:
            raise ValidationError(
                "Personal visibility cannot have any viewing groups assigned."
            )
        elif viewing_groups and visibility == TestJob.VISIBLE_PUBLIC:
            raise ValidationError(
                "Pulibc visibility cannot have any viewing groups assigned."
            )
        return self.cleaned_data["viewing_groups"]


class TestJobAdmin(admin.ModelAdmin):
    def requested_device_type_name(self, obj):
        return "" if obj.requested_device_type is None else obj.requested_device_type

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return settings.ALLOW_ADMIN_DELETE

    requested_device_type_name.short_description = "Request device type"
    form = VisibilityForm
    actions = [cancel_action, fail_action]
    list_filter = ("state", RequestedDeviceTypeFilter, ActualDeviceFilter)
    fieldsets = (
        (
            "Owner",
            {
                "fields": (
                    "user",
                    "group",
                    "submitter",
                    "is_public",
                    "visibility",
                    "viewing_groups",
                )
            },
        ),
        ("Request", {"fields": ("requested_device_type", "priority", "health_check")}),
        (
            "Advanced properties",
            {"fields": ("description", "tags", "sub_id", "target_group")},
        ),
        ("Current status", {"fields": ("actual_device", "state", "health")}),
        ("Results & Failures", {"fields": ("failure_tags", "failure_comment")}),
    )
    readonly_fields = ("state",)
    list_display = (
        "id",
        "state",
        "health",
        "submitter",
        "requested_device_type_name",
        "actual_device",
        "health_check",
        "submit_time",
        "start_time",
        "end_time",
    )
    ordering = ["-submit_time"]


def disable_health_check_action(
    modeladmin, request, queryset
):  # pylint: disable=unused-argument
    queryset.update(disable_health_check=False)


disable_health_check_action.short_description = "disable health checks"


class DeviceTypeAdmin(admin.ModelAdmin):
    def architecture_name(self, obj):
        if obj.architecture:
            return obj.architecture
        return ""

    architecture_name.short_description = "arch"

    def processor_name(self, obj):
        if obj.processor:
            return obj.processor
        return ""

    processor_name.short_description = "proc"

    def cpu_model_name(self, obj):
        if obj.cpu_model:
            return obj.cpu_model
        return ""

    cpu_model_name.short_description = "cpu"

    def bit_count(self, obj):
        if obj.bits:
            return obj.bits
        return ""

    bit_count.short_description = "bits"

    def list_of_cores(self, obj):
        if obj.core_count:
            return "%s x %s" % (
                obj.core_count,
                ",".join([core.name for core in obj.cores.all().order_by("name")]),
            )
        return ""

    list_of_cores.short_description = "cores"

    def get_readonly_fields(self, _, obj=None):
        if obj:  # editing an existing object
            return self.readonly_fields + ("name",)
        return self.readonly_fields

    def health_check_enabled(self, obj):
        return not obj.disable_health_check

    health_check_enabled.boolean = True
    health_check_enabled.short_description = "HC enabled"

    def health_check_frequency(self, device_type):
        if device_type.health_denominator == DeviceType.HEALTH_PER_JOB:
            return "every %d jobs" % device_type.health_frequency
        return "every %d hours" % device_type.health_frequency

    health_check_frequency.short_description = "HC frequency"

    actions = [disable_health_check_action]
    list_filter = ("name", "display", "cores", "architecture", "processor")
    list_display = (
        "name",
        "display",
        "owners_only",
        "health_check_enabled",
        "health_check_frequency",
        "architecture_name",
        "processor_name",
        "cpu_model_name",
        "list_of_cores",
        "bit_count",
    )
    fieldsets = (
        (
            "Properties",
            {"fields": ("name", "description", "display", "owners_only", "aliases")},
        ),
        (
            "Health checks",
            {
                "fields": (
                    ("health_frequency", "health_denominator"),
                    "disable_health_check",
                )
            },
        ),
        (
            "Meta data",
            {
                "fields": (
                    "architecture",
                    "processor",
                    ("cores", "core_count"),
                    "bits",
                    "cpu_model",
                )
            },
        ),
    )
    ordering = ["name"]


def worker_health_active(ModelAdmin, request, queryset):
    with transaction.atomic():
        for worker in queryset.select_for_update():
            worker.go_health_active(request.user)
            worker.save()


def worker_health_maintenance(ModelAdmin, request, queryset):
    with transaction.atomic():
        for worker in queryset.select_for_update():
            worker.go_health_maintenance(request.user)
            worker.save()


def worker_health_retired(ModelAdmin, request, queryset):
    with transaction.atomic():
        for worker in queryset.select_for_update():
            worker.go_health_retired(request.user)
            worker.save()


worker_health_active.short_description = "Update health of selected workers to Active"
worker_health_maintenance.short_description = (
    "Update health of selected workers to Maintenance"
)
worker_health_retired.short_description = "Update health of selected workers to Retired"


class WorkerAdmin(admin.ModelAdmin):
    list_display = ("hostname", "state", "health")
    readonly_fields = ("state",)
    ordering = ["hostname"]
    actions = [worker_health_active, worker_health_maintenance, worker_health_retired]

    def get_readonly_fields(self, _, obj=None):
        if obj:  # editing an existing object
            return self.readonly_fields + ("hostname",)
        return self.readonly_fields

    def has_delete_permission(self, request, obj=None):
        return settings.ALLOW_ADMIN_DELETE


class TagLowerForm(forms.ModelForm):
    def clean_name(self):
        name = self.cleaned_data["name"]
        if name != name.lower():
            raise ValidationError("Tag names are case-insensitive.")
        return name


class TagAdmin(admin.ModelAdmin):
    form = TagLowerForm
    list_display = ("name", "description")
    ordering = ["name"]

    def get_readonly_fields(self, _, obj=None):
        if obj:  # editing an existing object
            return self.readonly_fields + ("name",)
        return self.readonly_fields


class NotificationRecipientAdmin(admin.ModelAdmin):
    def handle(self, obj):
        if obj.method == NotificationRecipient.EMAIL:
            return obj.email_address
        else:
            return "%s@%s" % (obj.irc_handle, obj.irc_server_name)

    list_display = ("method", "handle", "status")
    list_filter = ("method", "status")

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return settings.ALLOW_ADMIN_DELETE


admin.site.register(Alias, AliasAdmin)
admin.site.register(Architecture, ArchitectureAdmin)
admin.site.register(BitWidth, BitWidthAdmin)
admin.site.register(Core, CoreAdmin)
admin.site.register(Device, DeviceAdmin)
admin.site.register(DeviceType, DeviceTypeAdmin)
admin.site.register(JobFailureTag)
admin.site.register(NotificationRecipient, NotificationRecipientAdmin)
admin.site.register(ProcessorFamily)
admin.site.register(TestJob, TestJobAdmin)
admin.site.register(Tag, TagAdmin)
admin.site.register(Worker, WorkerAdmin)



# -*- coding: utf-8 -*-
# Copyright (C) 2015-2018 Linaro Limited
#
# Author: Stevan Radakovic <stevan.radakovic@linaro.org>
#
# This file is part of LAVA.
#
# LAVA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# LAVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with LAVA.  If not, see <http://www.gnu.org/licenses/>.

import csv
import os
import shutil
import simplejson
import tempfile

from django.db import IntegrityError
from django.db.models import Q
from django.db.utils import ProgrammingError
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import Group
from django.core import serializers
from django.core.exceptions import PermissionDenied, FieldDoesNotExist, FieldError
from django.urls import reverse
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404, loader
from django.template import defaultfilters

from lava_server.bread_crumbs import BreadCrumb, BreadCrumbTrail

from lava_results_app.views import index
from lava_results_app.views.query.decorators import ownership_required
from lava_results_app.views.query.forms import QueryForm, QueryConditionForm

from lava_results_app.models import (
    Query,
    QueryGroup,
    QueryCondition,
    QueryOmitResult,
    QueryUpdatedError,
    TestCase,
    TestSuite,
    InvalidConditionsError,
    InvalidContentTypeError,
)

from lava_results_app.views.query.tables import (
    UserQueryTable,
    OtherQueryTable,
    GroupQueryTable,
    QueryTestJobTable,
    QueryTestCaseTable,
    QueryTestSuiteTable,
)

from lava_scheduler_app.models import TestJob

from django.contrib.contenttypes.models import ContentType
from django_tables2 import RequestConfig

from lava.utils.lavatable import LavaView


class QueryViewDoesNotExistError(Exception):
    """ Raise when corresponding query materialized view does not exist. """


class UserQueryView(LavaView):
    def get_queryset(self):
        return Query.objects.filter(
            is_archived=False, owner__id=self.request.user.id
        ).order_by("name")


class OtherQueryView(LavaView):
    def get_queryset(self):
        # All published queries which are not part
        # of any group.
        other_queries = (
            Query.objects.filter(is_archived=False, is_published=True, query_group=None)
            .exclude(owner__id=self.request.user.id)
            .order_by("name")
        )

        return other_queries


class GroupQueryView(LavaView):
    def __init__(self, request, group, **kwargs):
        super().__init__(request, **kwargs)
        self.query_group = group

    def get_queryset(self):
        # Specific group queries.
        group_queries = (
            Query.objects.filter(
                is_archived=False, is_published=True, query_group=self.query_group
            )
            .exclude(owner__id=self.request.user.id)
            .order_by("name")
        )

        return group_queries


QUERY_CONTENT_TYPE_TABLE = {
    TestJob: QueryTestJobTable,
    TestCase: QueryTestCaseTable,
    TestSuite: QueryTestSuiteTable,
}


class QueryCustomResultView(LavaView):
    def __init__(self, content_type, conditions, request, **kwargs):
        self.content_type = content_type
        self.conditions = conditions
        self.request = request
        super().__init__(request, **kwargs)

    def get_queryset(self):
        return Query.get_queryset(self.content_type, self.conditions).visible_by_user(
            self.request.user
        )


class QueryResultView(LavaView):
    def __init__(self, query, request, **kwargs):
        self.query = query
        self.request = request
        super().__init__(request, **kwargs)

    def get_queryset(self):
        return self.query.get_results(self.request.user)


@BreadCrumb("Queries", parent=index)
def query_list(request):

    group_tables = {}
    terms_data = search_data = discrete_data = {}
    for group in QueryGroup.objects.all():
        if group.query_set.count():
            prefix = "group_%s_" % group.id
            group_view = GroupQueryView(
                request, group, model=Query, table_class=GroupQueryTable
            )
            table = GroupQueryTable(group_view.get_table_data(prefix), prefix=prefix)
            search_data.update(table.prepare_search_data(group_view))
            discrete_data.update(table.prepare_discrete_data(group_view))
            terms_data.update(table.prepare_terms_data(group_view))
            group_tables[group.name] = table
            config = RequestConfig(request, paginate={"per_page": table.length})
            config.configure(table)

    prefix = "other_"
    other_view = OtherQueryView(request, model=Query, table_class=OtherQueryTable)
    other_query_table = OtherQueryTable(
        other_view.get_table_data(prefix), prefix=prefix
    )
    config = RequestConfig(request, paginate={"per_page": other_query_table.length})
    config.configure(other_query_table)
    search_data.update(other_query_table.prepare_search_data(other_view))
    discrete_data.update(other_query_table.prepare_discrete_data(other_view))
    terms_data.update(other_query_table.prepare_terms_data(other_view))

    prefix = "user_"
    view = UserQueryView(request, model=Query, table_class=UserQueryTable)
    user_query_table = UserQueryTable(view.get_table_data(prefix), prefix=prefix)
    config = RequestConfig(request, paginate={"per_page": user_query_table.length})
    config.configure(user_query_table)
    search_data.update(user_query_table.prepare_search_data(view))
    discrete_data.update(user_query_table.prepare_discrete_data(view))
    terms_data.update(user_query_table.prepare_terms_data(view))

    template = loader.get_template("lava_results_app/query_list.html")
    return HttpResponse(
        template.render(
            {
                "user_query_table": user_query_table,
                "other_query_table": other_query_table,
                "search_data": search_data,
                "discrete_data": discrete_data,
                "terms_data": terms_data,
                "group_tables": group_tables,
                "bread_crumb_trail": BreadCrumbTrail.leading_to(query_list),
                "context_help": ["lava-queries-charts"],
            },
            request=request,
        )
    )


@BreadCrumb("Query ~{username}/{name}", parent=query_list, needs=["username", "name"])
def query_display(request, username, name):

    query = get_object_or_404(Query, owner__username=username, name=name)

    if not request.user.is_superuser:
        if not query.is_published and query.owner != request.user:
            raise PermissionDenied

    view = QueryResultView(
        query=query,
        request=request,
        model=query.content_type.model_class(),
        table_class=QUERY_CONTENT_TYPE_TABLE[query.content_type.model_class()],
    )

    table = QUERY_CONTENT_TYPE_TABLE[query.content_type.model_class()](
        query, request.user, view.get_table_data()
    )

    try:
        config = RequestConfig(request, paginate={"per_page": table.length})
        config.configure(table)
    except ProgrammingError:
        raise QueryViewDoesNotExistError(
            "Query view does not exist. Please contact query owner or system "
            "administrator."
        )

    omitted = [
        result.content_object for result in QueryOmitResult.objects.filter(query=query)
    ]
    template = loader.get_template("lava_results_app/query_display.html")
    return HttpResponse(
        template.render(
            {
                "query": query,
                "entity": query.content_type.model,
                "conditions": Query.serialize_conditions(
                    query.querycondition_set.all()
                ),
                "omitted": omitted,
                "query_table": table,
                "terms_data": table.prepare_terms_data(view),
                "search_data": table.prepare_search_data(view),
                "discrete_data": table.prepare_discrete_data(view),
                "bread_crumb_trail": BreadCrumbTrail.leading_to(
                    query_display, username=username, name=name
                ),
                "context_help": ["lava-queries-charts"],
            },
            request=request,
        )
    )


@BreadCrumb("Custom Query", parent=query_list)
@login_required
def query_custom(request):

    try:
        content_type = Query.get_content_type(request.GET.get("entity"))
    except InvalidContentTypeError as e:
        messages.error(request, e)
        raise Http404()

    if content_type.model_class() not in QueryCondition.RELATION_MAP:
        messages.error(
            request, "Wrong table name in entity param. Please refer to query docs."
        )
        raise Http404()

    try:
        conditions = Query.parse_conditions(content_type, request.GET.get("conditions"))
    except InvalidConditionsError as e:
        messages.error(request, e)
        raise Http404()

    view = QueryCustomResultView(
        content_type=content_type,
        conditions=conditions,
        request=request,
        model=content_type.model_class(),
        table_class=QUERY_CONTENT_TYPE_TABLE[content_type.model_class()],
    )

    try:
        table = QUERY_CONTENT_TYPE_TABLE[content_type.model_class()](
            None, request.user, view.get_table_data()
        )
    except (FieldDoesNotExist, FieldError) as e:
        messages.error(request, e)
        raise Http404()

    config = RequestConfig(request, paginate={"per_page": table.length})
    config.configure(table)
    template = loader.get_template("lava_results_app/query_custom.html")
    return HttpResponse(
        template.render(
            {
                "query_table": table,
                "conditions": conditions,
                "terms_data": table.prepare_terms_data(view),
                "search_data": table.prepare_search_data(view),
                "discrete_data": table.prepare_discrete_data(view),
                "bread_crumb_trail": BreadCrumbTrail.leading_to(query_custom),
                "context_help": ["lava-queries-charts"],
            },
            request=request,
        )
    )


@BreadCrumb("Query ~{username}/{name}", parent=query_list, needs=["username", "name"])
@login_required
def query_detail(request, username, name):

    query = get_object_or_404(Query, owner__username=username, name=name)
    query_conditions = Query.serialize_conditions(query.querycondition_set.all())
    template = loader.get_template("lava_results_app/query_detail.html")
    return HttpResponse(
        template.render(
            {
                "query": query,
                "query_conditions": query_conditions,
                "bread_crumb_trail": BreadCrumbTrail.leading_to(
                    query_detail, username=username, name=name
                ),
                "context_help": ["lava-queries-charts"],
                "condition_form": QueryConditionForm(
                    instance=None, initial={"query": query, "table": query.content_type}
                ),
            },
            request=request,
        )
    )


@BreadCrumb("Add new", parent=query_list)
@login_required
def query_add(request):

    query = Query()
    query.owner = request.user

    if request.GET.get("entity"):
        query.content_type = Query.get_content_type(request.GET.get("entity"))

    return query_form(request, BreadCrumbTrail.leading_to(query_add), instance=query)


@BreadCrumb("Edit ~{username}/{name}", parent=query_detail, needs=["username", "name"])
@login_required
@ownership_required
def query_edit(request, username, name):

    query = get_object_or_404(Query, owner__username=username, name=name)

    return query_form(
        request,
        BreadCrumbTrail.leading_to(query_edit, username=username, name=name),
        instance=query,
    )


@login_required
@ownership_required
def query_delete(request, username, name):

    query = get_object_or_404(Query, owner__username=username, name=name)
    query.is_archived = True
    query.save()
    return HttpResponseRedirect(reverse("lava.results.query_list"))


@login_required
def query_export(request, username, name):
    """
    Create and serve the CSV file.
    """
    query = get_object_or_404(Query, owner__username=username, name=name)

    results = query.get_results(request.user)
    filename = "query_%s_%s_export" % (query.owner.username, query.name)
    return _export_query(results, query.content_type, filename)


@login_required
def query_export_custom(request):
    """
    Create and serve the CSV file.
    """

    try:
        content_type = Query.get_content_type(request.GET.get("entity"))
    except InvalidContentTypeError as e:
        messages.error(request, e)
        raise Http404()

    if content_type.model_class() not in QueryCondition.RELATION_MAP:
        messages.error(
            request, "Wrong table name in entity param. Please refer to query docs."
        )
        raise Http404()

    try:
        conditions = Query.parse_conditions(content_type, request.GET.get("conditions"))
    except InvalidConditionsError as e:
        messages.error(request, e)
        raise Http404()

    filename = "query_%s_export" % (content_type)

    try:
        results = Query.get_queryset(content_type, conditions).visible_by_user(
            request.user
        )
    except (FieldDoesNotExist, FieldError) as e:
        messages.error(request, e)
        raise Http404()

    return _export_query(results, content_type, filename)


@login_required
@ownership_required
def query_toggle_published(request, username, name):

    query = get_object_or_404(Query, owner__username=username, name=name)

    query.is_published = not query.is_published
    query.save()

    return HttpResponseRedirect(query.get_absolute_url() + "/+detail")


@BreadCrumb("Copy", parent=query_detail, needs=["username", "name"])
@login_required
def query_copy(request, username, name):
    query = get_object_or_404(Query, owner__username=username, name=name)
    query.owner = request.user
    query.is_published = False

    return query_form(
        request,
        BreadCrumbTrail.leading_to(query_copy, name=name, username=username),
        instance=query,
        is_copy=True,
    )


@login_required
@ownership_required
def query_refresh(request, name, username):

    query = get_object_or_404(Query, owner__username=username, name=name)

    success = True
    error_msg = None
    try:
        query.refresh_view()
    except QueryUpdatedError as e:
        error_msg = str(e)
        success = False
    except Exception as e:
        success = False
        error_msg = "%s<br>Please contact system administrator." % str(e)

    last_updated = defaultfilters.date(query.last_updated, "DATETIME_FORMAT")
    return HttpResponse(
        simplejson.dumps([success, str(last_updated), error_msg]),
        content_type="application/json",
    )


@login_required
def query_group_list(request):

    term = request.GET["term"]
    groups = [
        str(group.name) for group in QueryGroup.objects.filter(name__istartswith=term)
    ]
    return HttpResponse(simplejson.dumps(groups), content_type="application/json")


@login_required
def query_add_group(request, username, name):

    if request.method != "POST":
        raise PermissionDenied

    group_name = request.POST.get("value")
    query = get_object_or_404(Query, owner__username=username, name=name)
    old_group = query.query_group

    if not group_name:
        query.query_group = None
    else:
        new_group = QueryGroup.objects.get_or_create(name=group_name)[0]
        query.query_group = new_group

    query.save()

    if old_group:
        if not old_group.query_set.count():
            old_group.delete()

    return HttpResponseRedirect(query.get_absolute_url() + "/+detail")


@login_required
@ownership_required
def query_select_group(request, username, name):

    if request.method != "POST":
        raise PermissionDenied

    group_name = request.POST.get("value")
    query = get_object_or_404(Query, owner__username=username, name=name)

    if not group_name:
        query.group = None
    else:
        try:
            group = Group.objects.get(name=group_name)
        except Group.DoesNotExist:
            group = None
        query.group = group

    query.save()

    return HttpResponseRedirect(query.get_absolute_url() + "/+detail")


@login_required
def get_query_group_names(request):

    term = request.GET["term"]
    groups = []
    for group in Group.objects.filter(user=request.user, name__istartswith=term):
        groups.append({"id": group.id, "name": group.name, "label": group.name})
    return HttpResponse(simplejson.dumps(groups), content_type="application/json")


@login_required
@ownership_required
def query_add_condition(request, username, name):

    query = get_object_or_404(Query, owner__username=username, name=name)

    return query_condition_form(request, query, BreadCrumbTrail.leading_to(query_add))


@login_required
@ownership_required
def query_edit_condition(request, username, name, id):

    query = get_object_or_404(Query, owner__username=username, name=name)
    query_condition = get_object_or_404(QueryCondition, id=id)

    return query_condition_form(
        request, query, BreadCrumbTrail.leading_to(query_add), instance=query_condition
    )


@login_required
@ownership_required
def query_remove_condition(request, username, name, id):

    query = get_object_or_404(Query, owner__username=username, name=name)
    query_condition = get_object_or_404(QueryCondition, id=id)
    query_condition.delete()

    return HttpResponseRedirect(query.get_absolute_url() + "/+detail")


@login_required
@ownership_required
def query_omit_result(request, username, name, id):

    query = get_object_or_404(Query, owner__username=username, name=name)
    result_object = get_object_or_404(query.content_type.model_class(), id=id)

    try:
        QueryOmitResult.objects.create(query=query, content_object=result_object)
    except IntegrityError:
        # Ignore unique constraint violation.
        pass

    return HttpResponseRedirect(query.get_absolute_url())


@login_required
@ownership_required
def query_include_result(request, username, name, id):

    query = get_object_or_404(Query, owner__username=username, name=name)
    result_object = get_object_or_404(query.content_type.model_class(), id=id)

    try:
        QueryOmitResult.objects.get(
            query=query, object_id=result_object.id, content_type=query.content_type
        ).delete()
    except QueryOmitResult.DoesNotExist:
        # Ignore does not exist violation.
        raise

    return HttpResponseRedirect(query.get_absolute_url())


def get_query_names(request):

    term = request.GET["term"]
    result = []

    query_list = (
        Query.objects.filter(
            Q(is_archived=False),
            Q(name__istartswith=term) | Q(owner__username__istartswith=term),
        )
        .distinct()
        .order_by("name")
    )
    for query in query_list:
        result.append(
            {
                "value": query.owner_name,
                "label": query.owner_name,
                "id": query.id,
                "content_type": query.content_type.model_class().__name__,
            }
        )
    return HttpResponse(simplejson.dumps(list(result)), content_type="application/json")


def query_form(request, bread_crumb_trail, instance=None, is_copy=False):

    if request.method == "POST":

        form = QueryForm(request.user, request.POST, instance=instance, is_copy=is_copy)
        if form.is_valid():
            query = form.save()
            if request.GET.get("conditions"):
                conditions = Query.parse_conditions(
                    query.content_type, request.GET.get("conditions")
                )
                for condition in conditions:
                    condition.query = query
                    condition.save()

            return HttpResponseRedirect(query.get_absolute_url() + "/+detail")

    else:
        form = QueryForm(request.user, instance=instance, is_copy=is_copy)
        form.fields["owner"].initial = request.user

    query_name = None
    if is_copy:
        query_name = instance.name
        instance.name = None
    template = loader.get_template("lava_results_app/query_form.html")
    return HttpResponse(
        template.render(
            {
                "bread_crumb_trail": bread_crumb_trail,
                "is_copy": is_copy,
                "query_name": query_name,
                "form": form,
                "context_help": ["lava-queries-charts"],
            },
            request=request,
        )
    )


def query_condition_form(request, query, bread_crumb_trail, instance=None):

    form = QueryConditionForm(request.POST, instance=instance)

    if form.is_valid():
        query_condition = form.save()

        return HttpResponse(
            serializers.serialize("json", [query_condition, query_condition.table]),
            content_type="application/json",
        )
    else:
        return HttpResponse(
            simplejson.dumps(["fail", form.errors]), content_type="application/json"
        )


def _remove_dir(path):
    """ Removes directory @path. Doesn't raise exceptions. """
    try:
        # Delete directory.
        shutil.rmtree(path)
    except OSError:
        # Silent exception whatever happens. If it's unexisting dir, we don't
        # care.
        pass


def _export_query(query_results, content_type, filename):

    tmp_dir = tempfile.mkdtemp()
    file_path = os.path.join(tmp_dir, "%s.csv" % filename)

    query_keys = [field.name for field in content_type.model_class()._meta.get_fields()]

    query_keys.sort()
    # Remove non-relevant columns for CSV file.
    removed_fields = [
        # TestJob fields:
        "user_id",
        "actual_device_id",
        "definition",
        "admin_notifications",
        "group_id",
        "multinode_definition",
        "original_definition",
        "health_check",
        "sub_id",
        "submitter_id",
        "testdata",
        "testsuite",
        "target_group",
        "notification",
        "testjobmaterializedview",
        "testjobuser",
        "pipeline_compatibility",
        # TestSuite fields:
        "job_id",
        # TestCase fields:
        "actionlevels",
        "suite_id",
        "test_set_id",
    ]
    for field in removed_fields:
        if field in query_keys:
            query_keys.remove(field)

    with open(file_path, "w+") as csv_file:
        out = csv.DictWriter(
            csv_file,
            quoting=csv.QUOTE_ALL,
            extrasaction="ignore",
            fieldnames=query_keys,
        )
        out.writeheader()

        for result in query_results:
            result_dict = result.__dict__.copy()
            # Encode the strings if necessary.
            out.writerow(
                {
                    k: (v.encode("utf8") if isinstance(v, str) else v)
                    for k, v in result_dict.items()
                }
            )

    response = HttpResponse(content_type="text/csv")
    response["Content-Disposition"] = "attachment; filename=%s.csv" % filename
    with open(file_path, "r") as csv_file:
        response.write(csv_file.read())

    _remove_dir(tmp_dir)
    return response



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  requires.py
#
#  Copyright 2018 Linaro
#  Author: Neil Williams <neil.williams@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import sys
import yaml
import argparse

#  Goals:
#  0: Convert a packaged list of dependencies and versions
#     to a distro-specific list of dependencies.
#  1: Remove need for requirements.txt using Pip syntax as this is
#     misleading.
#  2: output a list of binary package names for the requested distribution
#     and suite to pass to docker scripts and LXC unit test jobs.


def debian(args, depends):
    """
    Special knowledge about how the dependencies work
    for this specific distribution.

    For package building, recurse from backports into
    the parent suite.
    For package names (e.g. docker), require TWO separate
    calls, 1 for backports and one for the parent.
    """
    if args.unittests:
        unittests = []
        if args.names and depends[args.package]:
            for key, item in depends[args.package].items():
                if depends[args.package][key].get("unittests"):
                    unittests.append(item["name"])
            if unittests:
                print(" ".join(sorted(unittests)))
        return 0
    if args.names:
        msg = []
        backports = []
        if not depends.get(args.package):
            return 0
        for key, item in depends[args.package].items():
            if depends[args.package][key].get("unittests"):
                continue
            if args.suite.endswith("-backports"):
                backports.append(item["name"])
                continue
            msg.append(item["name"])
        if backports:
            print(" ".join(sorted(backports)))
        elif msg:
            print(" ".join(sorted(msg)))
        return 0
    if not depends[args.package]:
        return 0
    for item in depends[args.package].keys():
        if depends[args.package][item].get("unittests"):
            continue
        print("%s%s" % (item, depends[args.package][item].get("version", "")))


def load_depends(args, parent):
    req = os.path.join(
        os.path.dirname(__file__), "requirements", args.distribution, "%s.yaml" % parent
    )
    if not os.path.exists(req):
        msg = "Unsupported suite|distribution: %s %s\n\n" % (args.distribution, parent)
        sys.stderr.write(msg)
        raise RuntimeError(msg)
    with open(req, "r") as data:
        depends = yaml.safe_load(data)
    if args.package not in depends:
        msg = "Unknown package: %s\n\n" % args.package
        sys.stderr.write(msg)
        raise RuntimeError(msg)
    return depends


def main():
    """
    Parse options and load requirements files.
    By default, outputs the same list as requirements.txt without
    packages needed for unit tests.
    Use the -n option to only get the distro|suite package names.
    Use the -u option to get the extra packages neeed for unit tests.
    """
    parser = argparse.ArgumentParser(description="Handle dependency lists")
    parser.add_argument(
        "-p", "--package", required=True, help="Name of the LAVA package."
    )
    parser.add_argument(
        "-d", "--distribution", required=True, help="Only Debian is supported currently"
    )
    parser.add_argument(
        "-s", "--suite", required=True, help="The distribution suite / release"
    )
    parser.add_argument(
        "-n", "--names", action="store_true", help="List the distribution package names"
    )
    parser.add_argument(
        "-u",
        "--unittests",
        action="store_true",
        help="Distribution package names for unittest support - requires --names",
    )
    args = parser.parse_args()
    if args.unittests and not args.names:
        raise RuntimeError("--unittests option requires --names")
    try:
        depends = load_depends(args, args.suite)
    except RuntimeError:
        return 1
    if args.distribution == "debian":
        debian(args, depends)
        if args.suite.endswith("-backports") and not args.names:
            parent = args.suite.replace("-backports", "")
            try:
                ret = load_depends(args, parent)
            except RuntimeError:
                return 2
            debian(args, ret)
    return 0


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018 Linaro Limited
#
# Author: Rémi Duraffort <remi.duraffort@linaro.org>
#
# This file is part of LAVA.
#
# LAVA is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

from voluptuous import Msg, Required

from lava_common.schemas import boot


def schema(live=False):
    base = {Required("method"): Msg("dfu", "'method' should be 'dfu'")}
    return {**boot.schema(live), **base}



# -*- coding: utf-8 -*-
# Copyright (C) 2017-2018 Linaro Limited
#
# Author: Remi Duraffort <remi.duraffort@linaro.org>
#
# This file is part of LAVA.
#
# LAVA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# LAVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with LAVA.  If not, see <http://www.gnu.org/licenses/>.

import os
from datetime import datetime, timedelta
import pytz
import xmlrpc.client
from django.utils import timezone
from linaro_django_xmlrpc.models import ExposedV2API
from lava_scheduler_app.api import SchedulerAPI
from lava_scheduler_app.logutils import read_logs
from lava_scheduler_app.models import TestJob
from lava_results_app.models import TestCase


def load_optional_file(filename):
    try:
        with open(filename, "r") as f_in:
            return f_in.read().encode("utf-8")
    except OSError:
        return None


class SchedulerJobsAPI(ExposedV2API):
    def cancel(self, job_id):
        """
        Name
        ----
        `scheduler.jobs.cancel` (`job_id`)

        Description
        -----------
        Cancel the given job referred by its id.

        Arguments
        ---------
        `job_id`: string
            Job id which should be canceled.

        Return value
        ------------
        None. The user should be authenticated with an username and token.
        """
        cls = SchedulerAPI(self._context)
        return cls.cancel_job(job_id)

    def configuration(self, job_id):
        """
        Name
        ----
        `scheduler.jobs.configuration` (`job_id`)

        Description
        -----------
        Return the full job configuration

        Arguments
        ---------
        `job_id`: string
            Job id

        Return value
        ------------
        Return an array with [job, device, dispatcher, env, env-dut] config.
        Any of theses values might be None if the corresponding file hasn't
        been used by the job.
        If the job hasn't started yet, a 404 error will be returned.
        """
        try:
            job = TestJob.get_by_job_number(job_id)
        except TestJob.DoesNotExist:
            raise xmlrpc.client.Fault(404, "Job '%s' was not found." % job_id)

        if not job.can_view(self.user):
            raise xmlrpc.client.Fault(
                403, "Job '%s' not available to user '%s'." % (job_id, self.user)
            )

        if job.state not in [
            TestJob.STATE_RUNNING,
            TestJob.STATE_CANCELING,
            TestJob.STATE_FINISHED,
        ]:
            raise xmlrpc.client.Fault(404, "Job '%s' has not started yet" % job_id)

        output_dir = job.output_dir
        definition = load_optional_file(os.path.join(output_dir, "job.yaml"))
        device = load_optional_file(os.path.join(output_dir, "device.yaml"))
        dispatcher = load_optional_file(os.path.join(output_dir, "dispatcher.yaml"))
        env = load_optional_file(os.path.join(output_dir, "env.yaml"))
        env_dut = load_optional_file(os.path.join(output_dir, "env.dut.yaml"))
        return [definition, device, dispatcher, env, env_dut]

    def definition(self, job_id):
        """
        Name
        ----
        `scheduler.jobs.definition` (`job_id`)

        Description
        -----------
        Return the job definition

        Arguments
        ---------
        `job_id`: string
            Job id

        Return value
        ------------
        The job definition or and error.

        Note: for MultiNode jobs, the original MultiNode definition
        is returned.
        """
        try:
            job = TestJob.get_by_job_number(job_id)
        except TestJob.DoesNotExist:
            raise xmlrpc.client.Fault(404, "Job '%s' was not found." % job_id)

        if not job.can_view(self.user):
            raise xmlrpc.client.Fault(
                403, "Job '%s' not available to user '%s'." % (job_id, self.user)
            )

        if job.is_multinode:
            return job.multinode_definition
        return job.original_definition

    def list(self, state=None, health=None, start=0, limit=25, since=0, verbose=False):
        """
        Name
        ----
        `scheduler.jobs.list` (
            `state=None`, `health=None`, `start=0`, `limit=25`,
            `since=None`, `verbose=False`
        )

        Description
        -----------
        List the last jobs, within the specified range, in descending order of
        job ID. Jobs can be filtered by `state` and `health` (if provided).

        Arguments
        ---------
        `state`: str
          Filter by state, None by default (no filtering).
          Values: [SUBMITTED, SCHEDULING, SCHEDULED, RUNNING, CANCELING, FINISHED]
        `health`: str
          Filter by health, None by default (no filtering).
          Values: [UNKNOWN, COMPLETE, INCOMPLETE, CANCELED]
        `start`: int
          Skip the first N job(s) in the list
        `limit`: int
          Max number of jobs to return.
          This value will be clamped to 100
        `since`: int (minutes)
          Filter by jobs which completed in the last N minutes.
        `verbose`: bool
          Add extra data including actual_device, start_time, end_time,
          error_msg and error_type.
          Note: error_msg can contain nested quotes and other escape
          characters, parse with care.

        Return value
        ------------
        This function returns an array of jobs with keys:
            "id", "description", "device_type", "health",
            "state", "submitter"
        If verbose is True, these keys are added:
            "actual_device", "start_time", "end_time",
            "error_msg", "error_type"
        """
        ret = []
        start = max(0, start)
        limit = min(limit, 100)
        jobs = TestJob.objects.all().select_related(
            "requested_device_type", "submitter"
        )
        if state:
            try:
                jobs = jobs.filter(state=TestJob.STATE_REVERSE[state.capitalize()])
            except KeyError:
                raise xmlrpc.client.Fault(400, "Invalid state '%s'" % state)
        if health:
            try:
                jobs = jobs.filter(health=TestJob.HEALTH_REVERSE[health.capitalize()])
            except KeyError:
                raise xmlrpc.client.Fault(400, "Invalid health '%s'" % health)

        # since
        if since:
            end_time = datetime.now()
            # search back in time
            start_time = end_time - timedelta(minutes=since)
            try:
                jobs = jobs.filter(end_time__range=[start_time, end_time])
            except TestJob.DoesNotExist:
                raise xmlrpc.client.Fault(
                    404, "No jobs exist since %s minutes ago" % since
                )

        if not jobs:
            raise xmlrpc.client.Fault(404, "No jobs match the specified criteria.")

        for job in jobs.order_by("-id")[start : start + limit]:
            device_type = None
            if job.requested_device_type is not None:
                device_type = job.requested_device_type.name
            data = {
                "id": job.display_id,
                "description": job.description,
                "device_type": device_type,
                "health": job.get_health_display(),
                "state": job.get_state_display(),
                "submitter": job.submitter.username,
            }
            if verbose:
                actual_device = None
                # Neither dynamic connections
                # nor jobs cancelled in submitted state
                # will have an actual_device
                if job.actual_device:
                    actual_device = job.actual_device.hostname
                # cancelled jobs might not have start or end time
                end_time = str(job.end_time) if job.end_time else None
                start_time = str(job.start_time) if job.start_time else None
                metadata = None
                try:
                    job_case = TestCase.objects.get(
                        suite__job=job, suite__name="lava", name="job"
                    )
                except TestCase.DoesNotExist:
                    job_case = None
                if job_case:
                    metadata = job_case.action_metadata
                if not metadata:
                    # if job_case exists but metadata is still None due to job error
                    metadata = {}
                data.update(
                    {
                        "actual_device": actual_device,
                        "start_time": start_time,
                        "end_time": end_time,
                        "error_msg": metadata.get("error_msg"),
                        "error_type": metadata.get("error_type"),
                    }
                )

            ret.append(data)

        return ret

    def queue(self, device_types=None, start=0, limit=25):
        """
        Name
        ----
        `scheduler.jobs.queue` (`device_types=None`, `start=0`, `limit=25`)

        Description
        -----------
        List the queued jobs (state.SUBMITTED), within the specified range,
        in descending order of job ID.
        Jobs can be filtered by `requested_device_type` (if provided).

        Arguments
        ---------
        `device_types`: Array of str
          If provided, list jobs whose requested_device_type match any of the
          provided device-types. None by default (no filtering).
        `start`: int
          Skip the first N job(s) in the list
        `limit`: int
          Max number of jobs to return.
          This value will be clamped to 100

        Return value
        ------------
        This function returns an array of jobs with keys:
            "id", "description", "requested_device_type", "submitter"

        If no queued test jobs exist to match the criteria, an empty array
        is returned.
        """
        ret = []
        start = max(0, start)
        limit = min(limit, 100)
        jobs = TestJob.objects.filter(state=TestJob.STATE_SUBMITTED).select_related(
            "requested_device_type", "submitter"
        )
        if device_types is not None:
            jobs = jobs.filter(requested_device_type__name__in=device_types)

        for job in jobs.order_by("-id")[start : start + limit]:
            data = {
                "id": job.display_id,
                "description": job.description,
                "requested_device_type": job.requested_device_type.name,
                "submitter": job.submitter.username,
            }
            ret.append(data)

        return ret

    def logs(self, job_id, start=0, end=None):
        """
        Name
        ----
        `scheduler.jobs.logs` (`job_id`, `start=0`, `end=None`)

        Description
        -----------
        Return the logs for the given job

        Arguments
        ---------
        `job_id`: str
          Job id
        `start`: int
          Show only after the given line
        `end`: int
          Do not return after the fiven line

        Return value
        ------------
        This function returns a tuple made of (job_finished, data).
        job_finished is True if and only if the job is finished.
        """
        try:
            job = TestJob.get_by_job_number(job_id)
        except TestJob.DoesNotExist:
            raise xmlrpc.client.Fault(404, "Job '%s' was not found." % job_id)

        if not job.can_view(self.user):
            raise xmlrpc.client.Fault(
                403, "Job '%s' not available to user '%s'." % (job_id, self.user)
            )

        job_finished = job.state == TestJob.STATE_FINISHED

        try:
            data = read_logs(job.output_dir, start, end)
            return (job_finished, xmlrpc.client.Binary(data.encode("utf-8")))
        except OSError:
            return (job_finished, xmlrpc.client.Binary("[]".encode("utf-8")))

    def show(self, job_id):
        """
        Name
        ----
        `scheduler.jobs.show` (`job_id`)

        Description
        -----------
        Show job details

        Arguments
        ---------
        `job_id`: string
          Job id

        Return value
        ------------
        This function returns a dictionary of details abou the specified test job.
        """
        try:
            job = TestJob.get_by_job_number(job_id)
        except TestJob.DoesNotExist:
            raise xmlrpc.client.Fault(404, "Job '%s' was not found." % job_id)

        if not job.can_view(self.user):
            raise xmlrpc.client.Fault(
                403, "Job '%s' not available to user '%s'." % (job_id, self.user)
            )

        device_hostname = None
        if job.actual_device is not None:
            device_hostname = job.actual_device.hostname

        device_type = None
        if job.requested_device_type is not None:
            device_type = job.requested_device_type.name

        return {
            "id": job.display_id,
            "description": job.description,
            "device": device_hostname,
            "device_type": device_type,
            "health_check": job.health_check,
            "pipeline": True,
            "health": job.get_health_display(),
            "state": job.get_state_display(),
            "submitter": job.submitter.username,
            "submit_time": job.submit_time,
            "start_time": job.start_time,
            "end_time": job.end_time,
            "tags": [t.name for t in job.tags.all()],
            "visibility": job.get_visibility_display(),
            "failure_comment": job.failure_comment,
        }

    def resubmit(self, job_id):
        """
        Name
        ----
        `scheduler.jobs.resubmit` (`job_id`)

        Description
        -----------
        Resubmit the given job referred by its id.

        Arguments
        ---------
        `job_id`: string
            The job's id which should be re-submitted.

        Return value
        ------------
        This function returns an XML-RPC integer which is the newly created
        job's id, provided the user is authenticated with an username and token.
        """
        cls = SchedulerAPI(self._context)
        return cls.resubmit_job(job_id)

    def submit(self, definition):
        """
        Name
        ----
        `scheduler.jobs.submit` (`definition`)

        Description
        -----------
        Submit the given job data which is in LAVA job JSON or YAML format as a
        new job to LAVA scheduler.

        Arguments
        ---------
        `definition`: string
            Job JSON or YAML string.

        Return value
        ------------
        This function returns an XML-RPC integer which is the newly created
        job's id, provided the user is authenticated with an username and token.
        If the job is a multinode job, this function returns the list of created
        job IDs.
        """
        cls = SchedulerAPI(self._context)
        return cls.submit_job(definition)



# Copyright (C) 2010, 2011 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#
# This file is part of linaro-django-xmlrpc.
#
# linaro-django-xmlrpc is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# linaro-django-xmlrpc is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with linaro-django-xmlrpc.  If not, see <http://www.gnu.org/licenses/>.

"""
XML-RPC views
"""

import base64

from django.contrib.auth.decorators import login_required
from django.contrib.sites.models import Site
from django.urls import reverse
from django.conf import settings
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404, redirect, render
from django.views.decorators.csrf import csrf_exempt

from linaro_django_xmlrpc.models import AuthToken, CallContext, Dispatcher, SystemAPI
from linaro_django_xmlrpc.forms import AuthTokenForm

from lava_server.bread_crumbs import BreadCrumb, BreadCrumbTrail
from lava_server.views import index as lava_index


@csrf_exempt
def handler(request, mapper, help_view):  # pylint: disable=too-many-return-statements
    """
    XML-RPC handler.

    If post data is defined, it assumes it's XML-RPC and tries to process as
    such. Empty POST request and GET requests assumes you're viewing from a
    browser and tells you about the service by redirecting you to a dedicated
    help page. For backwards compatibility the help view defaults to the
    'default_help' that shows what is registered in the global mapper. If you
    want to show help specific to your mapper you must specify help_view. It
    accepts whatever django.shortcuts.redirect() would.
    """
    if len(request.body):
        raw_data = request.body
        dispatcher = Dispatcher(mapper)

        auth_string = request.META.get("HTTP_AUTHORIZATION")

        if auth_string is not None:
            if " " not in auth_string:
                return HttpResponse("Invalid HTTP_AUTHORIZATION header", status=400)
            scheme, value = auth_string.split(" ", 1)
            if scheme != "Basic":
                return HttpResponse(
                    "Unsupported HTTP_AUTHORIZATION header, only Basic scheme is supported",
                    status=400,
                )
            try:
                decoded_value = base64.standard_b64decode(value).decode("utf-8")
            except (TypeError, UnicodeDecodeError):
                return HttpResponse(
                    "Corrupted HTTP_AUTHORIZATION header, bad base64 encoding",
                    status=400,
                )
            try:
                username, secret = decoded_value.split(":", 1)
            except ValueError:
                return HttpResponse(
                    "Corrupted HTTP_AUTHORIZATION header, no user:pass", status=400
                )
            user = AuthToken.get_user_for_secret(username, secret)
            if user is None:
                response = HttpResponse("Invalid token", status=401)
                response[
                    "WWW-Authenticate"
                ] = 'Basic realm="XML-RPC Authentication token"'
                return response
        else:
            user = request.user
        result = dispatcher.marshalled_dispatch(raw_data, user, request)
        response = HttpResponse(content_type="application/xml")
        response.write(result)
        response["Content-length"] = str(len(response.content))
        return response
    else:
        return redirect(help_view)


@BreadCrumb("API help", parent=lava_index)
def help(
    request, mapper, template_name="linaro_django_xmlrpc/api.html"
):  # pylint: disable=redefined-builtin
    context = CallContext(user=None, mapper=mapper, dispatcher=None)
    system = SystemAPI(context)
    if settings.HTTPS_XML_RPC:
        scheme = "https"
    else:
        scheme = request.META.get("REQUEST_SCHEME", "http")
    dashboard_methods = []
    scheduler_methods = []
    results_methods = []
    system_methods = []
    for method in system.listMethods():
        if "dashboard" in method:
            dashboard_methods.append(method)
        elif "scheduler" in method:
            scheduler_methods.append(method)
        elif "results" in method:
            results_methods.append(method)
        else:
            system_methods.append(method)
    methods = {
        "dashboard": [
            {
                "name": method,
                "signature": system.methodSignature(method),
                "help": system.methodHelp(method),
            }
            for method in dashboard_methods
        ],
        "scheduler": [
            {
                "name": method,
                "section": method.rsplit(".", 1)[0] if "." in method else "",
                "signature": system.methodSignature(method),
                "help": system.methodHelp(method),
            }
            for method in scheduler_methods
        ],
        "results": [
            {
                "name": method,
                "signature": system.methodSignature(method),
                "help": system.methodHelp(method),
            }
            for method in results_methods
        ],
        "system": [
            {
                "name": method,
                "signature": system.methodSignature(method),
                "help": system.methodHelp(method),
            }
            for method in system_methods
        ],
    }
    domain = Site.objects.get_current().domain
    return render(
        request,
        template_name,
        {
            "methods": methods,
            "context_help": ["data-export"],
            "bread_crumb_trail": BreadCrumbTrail.leading_to(help),
            "site_scheme": scheme,
            "site_domain": domain,
            "site_url": "{scheme}://{domain}".format(scheme=scheme, domain=domain),
        },
    )


@BreadCrumb("API tokens", parent=lava_index)
@login_required
def tokens(request):
    """
    List of tokens for an authenticated user
    """
    token_list = AuthToken.objects.filter(user=request.user).order_by("last_used_on")
    unused = AuthToken.objects.filter(
        user=request.user, last_used_on__isnull=True
    ).count()
    return render(
        request,
        "linaro_django_xmlrpc/tokens.html",
        {
            "token_list": token_list,
            "unused": unused,
            "bread_crumb_trail": BreadCrumbTrail.leading_to(tokens),
            "context_help": ["first_steps"],
        },
    )


@login_required
def create_token(request):
    """
    Create a token for the requesting user
    """
    if request.method == "POST":
        form = AuthTokenForm(request.POST)
        if form.is_valid():
            form.save(commit=False)
            form.instance.user = request.user
            form.instance.save()
            return HttpResponseRedirect(reverse("linaro_django_xmlrpc_tokens"))
    else:
        form = AuthTokenForm()
    return render(request, "linaro_django_xmlrpc/create_token.html", {"form": form})


@login_required
def delete_token(request, object_id):
    token = get_object_or_404(AuthToken, pk=object_id, user=request.user)
    if request.method == "POST":
        token.delete()
        return HttpResponseRedirect(reverse("linaro_django_xmlrpc_tokens"))
    return render(
        request, "linaro_django_xmlrpc/authtoken_confirm_delete.html", {"token": token}
    )


@login_required
def edit_token(request, object_id):
    token = get_object_or_404(AuthToken, pk=object_id, user=request.user)
    if request.method == "POST":
        form = AuthTokenForm(request.POST, instance=token)
        if form.is_valid():
            form.save()
            return HttpResponseRedirect(reverse("linaro_django_xmlrpc_tokens"))
    else:
        form = AuthTokenForm(instance=token)
    return render(
        request, "linaro_django_xmlrpc/edit_token.html", {"token": token, "form": form}
    )


@login_required
def delete_unused_tokens(request):
    token_list = AuthToken.objects.filter(user=request.user, last_used_on__isnull=True)
    if request.method == "POST":
        for token in token_list:
            token.delete()
        return HttpResponseRedirect(reverse("linaro_django_xmlrpc_tokens"))
    return render(
        request,
        "linaro_django_xmlrpc/tokens.html",
        {"token_list": token_list, "context_help": ["lava-tool"]},
    )



# Copyright (C) 2014 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

import os
import yaml
import datetime
from lava_dispatcher.action import Action, Pipeline
from lava_common.timeout import Timeout
from lava_common.exceptions import InfrastructureError, JobError
from lava_dispatcher.parser import JobParser
from lava_dispatcher.device import NewDevice
from lava_dispatcher.actions.deploy.testdef import get_test_action_namespaces
from lava_dispatcher.tests.utils import DummyLogger
from lava_dispatcher.job import Job
from lava_dispatcher.protocols.multinode import MultinodeProtocol
from lava_dispatcher.protocols.vland import VlandProtocol
from lava_dispatcher.tests.test_basic import Factory, StdoutTestCase
from lava_dispatcher.actions.test.shell import TestShellRetry, TestShellAction


# pylint: disable=duplicate-code,too-few-public-methods


class TestDefinitionHandlers(StdoutTestCase):  # pylint: disable=too-many-public-methods
    def setUp(self):
        super().setUp()
        self.factory = Factory()
        self.job = self.factory.create_kvm_job("sample_jobs/kvm.yaml")

    def test_testshell(self):
        testshell = None
        for action in self.job.pipeline.actions:
            self.assertIsNotNone(action.name)
            if isinstance(action, TestShellRetry):
                testshell = action.pipeline.actions[0]
                break
        self.assertIsInstance(testshell, TestShellAction)
        self.assertTrue(testshell.valid)

        if "timeout" in testshell.parameters:
            time_int = Timeout.parse(testshell.parameters["timeout"])
        else:
            time_int = Timeout.default_duration()
        self.assertEqual(
            datetime.timedelta(seconds=time_int).total_seconds(),
            testshell.timeout.duration,
        )

    def test_missing_handler(self):
        (rendered, _) = self.factory.create_device("kvm01.jinja2")
        device = NewDevice(yaml.safe_load(rendered))
        kvm_yaml = os.path.join(os.path.dirname(__file__), "sample_jobs/kvm.yaml")
        parser = JobParser()
        with open(kvm_yaml) as sample_job_data:
            data = yaml.safe_load(sample_job_data)
        data["actions"][2]["test"]["definitions"][0]["from"] = "unusable-handler"
        try:
            job = parser.parse(yaml.dump(data), device, 4212, None, "")
            job.logger = DummyLogger()
        except JobError:
            pass
        except Exception as exc:  # pylint: disable=broad-except
            self.fail(exc)
        else:
            self.fail("JobError not raised")

    def test_eventpatterns(self):
        testshell = None
        for action in self.job.pipeline.actions:
            self.assertIsNotNone(action.name)
            if isinstance(action, TestShellRetry):
                testshell = action.pipeline.actions[0]
                break
        self.assertTrue(testshell.valid)
        self.assertFalse(testshell.check_patterns("exit", None, ""))
        self.assertRaises(
            InfrastructureError, testshell.check_patterns, "eof", None, ""
        )
        self.assertTrue(testshell.check_patterns("timeout", None, ""))


class X86Factory(Factory):
    def create_x86_job(self, filename, device, validate_job=True):
        return self.create_job(device, filename, validate_job=validate_job)


class TestMultiNodeOverlay(StdoutTestCase):  # pylint: disable=too-many-public-methods
    def setUp(self):
        super().setUp()
        factory = X86Factory()
        self.server_job = factory.create_x86_job(
            "sample_jobs/test_action-1.yaml",
            "lng-generator-01.jinja2",
            validate_job=False,
        )
        self.client_job = factory.create_x86_job(
            "sample_jobs/test_action-2.yaml",
            "lng-generator-02.jinja2",
            validate_job=False,
        )

    def test_action_namespaces(self):
        self.assertIsNotNone(self.server_job)
        self.assertIsNotNone(self.client_job)
        deploy_server = [
            action
            for action in self.server_job.pipeline.actions
            if action.name == "tftp-deploy"
        ][0]
        self.assertIn(MultinodeProtocol.name, deploy_server.parameters.keys())
        self.assertIn(VlandProtocol.name, deploy_server.parameters.keys())
        self.assertEqual(
            ["common"], get_test_action_namespaces(self.server_job.parameters)
        )
        namespace = self.server_job.parameters.get("namespace")
        self.assertIsNone(namespace)
        namespace = self.client_job.parameters.get("namespace")
        self.assertIsNone(namespace)
        deploy_client = [
            action
            for action in self.client_job.pipeline.actions
            if action.name == "tftp-deploy"
        ][0]
        self.assertIn(MultinodeProtocol.name, deploy_client.parameters.keys())
        self.assertIn(VlandProtocol.name, deploy_client.parameters.keys())
        key_list = []
        for block in self.client_job.parameters["actions"]:
            key_list.extend(block.keys())
        self.assertEqual(key_list, ["deploy", "boot", "test"])  # order is important
        self.assertEqual(
            ["common"], get_test_action_namespaces(self.client_job.parameters)
        )
        key_list = []
        for block in self.server_job.parameters["actions"]:
            key_list.extend(block.keys())
        self.assertEqual(key_list, ["deploy", "boot", "test"])  # order is important


class TestShellResults(StdoutTestCase):  # pylint: disable=too-many-public-methods
    class FakeJob(Job):
        pass

    class FakeDeploy:
        """
        Derived from object, *not* Deployment as this confuses python -m unittest discover
        - leads to the FakeDeploy being called instead.
        """

        def __init__(self, parent):
            self.__parameters__ = {}
            self.pipeline = parent
            self.job = parent.job
            self.action = TestShellResults.FakeAction()

    class FakePipeline(Pipeline):
        def __init__(self, parent=None, job=None):
            super().__init__(parent, job)

    class FakeAction(Action):
        """
        Isolated Action which can be used to generate artificial exceptions.
        """

        name = "fake-action"
        description = "fake, do not use outside unit tests"
        summary = "fake action for unit tests"

        def __init__(self):
            super().__init__()
            self.count = 1

        def run(self, connection, max_end_time):
            self.count += 1
            raise JobError("fake error")



# Copyright (C) 2016 Linaro Limited
#
# Author: Dean Arnold <dean.arnold@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

# List just the subclasses supported for this base strategy
# imported by the parser to populate the list of subclasses.

from lava_dispatcher.action import Pipeline
from lava_dispatcher.logical import Deployment
from lava_dispatcher.actions.deploy import DeployAction
from lava_dispatcher.actions.deploy.download import DownloaderAction
from lava_dispatcher.actions.deploy.apply_overlay import (
    ExtractNfsRootfs,
    OverlayAction,
    ExtractModules,
    ApplyOverlayTftp,
)
from lava_dispatcher.actions.deploy.environment import DeployDeviceEnvironment


class Nfs(Deployment):
    """
    Strategy class for a NFS deployment.
    Downloads rootfs and deploys to NFS server on dispatcher
    """

    compatibility = 1
    name = "nfs"

    def __init__(self, parent, parameters):
        super().__init__(parent)
        self.action = NfsAction()
        self.action.section = self.action_type
        self.action.job = self.job
        parent.add_action(self.action, parameters)

    @classmethod
    def accepts(cls, device, parameters):
        if "to" not in parameters:
            return False, '"to" is not in deploy parameters'
        if parameters["to"] != "nfs":
            return False, '"to" parameter is not "nfs"'
        if "image" in device["actions"]["deploy"]["methods"]:
            return False, '"image" was in the device configuration deploy methods'
        if "nfs" in device["actions"]["deploy"]["methods"]:
            return True, "accepted"
        return False, '"nfs" was not in the device configuration deploy methods"'


class NfsAction(DeployAction):  # pylint:disable=too-many-instance-attributes

    name = "nfs-deploy"
    description = "deploy nfsrootfs"
    summary = "NFS deployment"

    def validate(self):
        super().validate()
        if not self.valid:
            return
        if "nfsrootfs" in self.parameters and "persistent_nfs" in self.parameters:
            self.errors = "Only one of nfsrootfs or persistent_nfs can be specified"

    def populate(self, parameters):
        download_dir = self.mkdtemp()
        self.internal_pipeline = Pipeline(
            parent=self, job=self.job, parameters=parameters
        )
        if "nfsrootfs" in parameters:
            self.internal_pipeline.add_action(
                DownloaderAction("nfsrootfs", path=download_dir)
            )
        if "modules" in parameters:
            self.internal_pipeline.add_action(
                DownloaderAction("modules", path=download_dir)
            )
        # NfsAction is a deployment, so once the nfsrootfs has been deployed, just do the overlay
        self.internal_pipeline.add_action(ExtractNfsRootfs())
        self.internal_pipeline.add_action(OverlayAction())
        self.internal_pipeline.add_action(ExtractModules())
        self.internal_pipeline.add_action(ApplyOverlayTftp())
        if self.test_needs_deployment(parameters):
            self.internal_pipeline.add_action(DeployDeviceEnvironment())



# -*- coding: utf-8 -*-
# Copyright (C) 2015-2018 Linaro Limited
#
# Author: Remi Duraffort <remi.duraffort@linaro.org>
#
# This file is part of LAVA.
#
# LAVA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# LAVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with LAVA.  If not, see <http://www.gnu.org/licenses/>.

import os
import zmq
from zmq.utils.strtypes import u

from django.conf import settings

from lava_server.cmdutils import LAVADaemonCommand


TIMEOUT = 5
FORMAT = "%(asctime)-15s %(levelname)7s %(message)s"


class Command(LAVADaemonCommand):
    help = "LAVA event publisher"
    default_logfile = "/var/log/lava-server/lava-publisher.log"

    def forward_event(self, leaving):
        try:
            sockets = dict(self.poller.poll(TIMEOUT * 1000 if leaving else None))
        except zmq.error.ZMQError as exc:
            self.logger.error("[POLL] Received a ZMQ error: %s", exc)
            return True

        if sockets.get(self.pipe_r) == zmq.POLLIN:
            # remove the message from the queue
            os.read(self.pipe_r, 1)
            if leaving:
                self.logger.warning(
                    "[POLL] signal already handled, please wait for the process to exit"
                )
                return True
            else:
                self.logger.info("[POLL] received a signal, leaving")
                return False

        if sockets.get(self.pull) == zmq.POLLIN:
            msg = self.pull.recv_multipart()
            self.logger.debug("[POLL] Forwarding: %s", msg)
            self.pub.send_multipart(msg)
            for (i, sock) in enumerate(self.additional_sockets):
                # Send in non blocking mode and print an error message if
                # the HWM is reached (because the receiver does not grab
                # the messages fast enough).
                try:
                    sock.send_multipart(msg, flags=zmq.DONTWAIT)
                except zmq.error.Again:
                    self.logger.warning("[POLL] Fail to forward to socket %d", i)
        return not leaving

    def handle(self, *args, **options):
        self.setup_logging(
            "lava-publisher", options["level"], options["log_file"], FORMAT
        )

        self.logger.info("[INIT] Dropping privileges")
        if not self.drop_privileges(options["user"], options["group"]):
            self.logger.error("[INIT] Unable to drop privileges")
            return

        if not settings.EVENT_NOTIFICATION:
            self.logger.error(
                "[INIT] 'EVENT_NOTIFICATION' is set to False, "
                "LAVA won't generated any events"
            )

        self.logger.info(
            "[INIT] Creating the input socket at %s", settings.INTERNAL_EVENT_SOCKET
        )
        context = zmq.Context.instance()
        self.pull = context.socket(zmq.PULL)
        self.pull.bind(settings.INTERNAL_EVENT_SOCKET)
        self.poller = zmq.Poller()
        self.poller.register(self.pull, zmq.POLLIN)

        # Translate signals into zmq messages
        (self.pipe_r, _) = self.setup_zmq_signal_handler()
        self.poller.register(self.pipe_r, zmq.POLLIN)

        # Create the default publishing socket
        self.logger.info(
            "[INIT] Creating the publication socket at %s", settings.EVENT_SOCKET
        )
        self.pub = context.socket(zmq.PUB)
        self.pub.bind(settings.EVENT_SOCKET)
        # Create the additional PUSH sockets
        if settings.EVENT_ADDITIONAL_SOCKETS:
            self.logger.info("[INIT] Creating the additional sockets:")
        self.additional_sockets = []
        for url in settings.EVENT_ADDITIONAL_SOCKETS:
            self.logger.info("[INIT]  * %s", url)
            sock = context.socket(zmq.PUSH)
            # Allow zmq to keep 10000 pending messages in each queue
            sock.setsockopt(zmq.SNDHWM, 10000)
            sock.connect(url)
            self.additional_sockets.append(sock)

        self.logger.info("[INIT] Starting the proxy")
        while self.forward_event(False):
            pass

        # Carefully close the logging socket as we don't want to lose messages
        self.logger.info("[EXIT] Disconnect pull socket and process messages")
        endpoint = u(self.pull.getsockopt(zmq.LAST_ENDPOINT))
        self.logger.debug("[EXIT] unbinding from '%s'", endpoint)
        self.pull.unbind(endpoint)

        while self.forward_event(True):
            pass

        # Close the sockets allowing 1s each to leave
        self.logger.info("[EXIT] Closing the sockets: the queue is empty")
        self.pull.close(linger=1)
        self.pub.close(linger=1)
        for socket in self.additional_sockets:
            socket.close(linger=1)
        context.term()



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-10-18 14:34
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dashboard_app', '0022_alter_imagechartfilter'),
        ('lava_scheduler_app', '0029_remove_testjob__results_bundle'),
        ('lava_results_app', '0014_xaxis_maxlength_increase'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='attachment',
            name='content_type',
        ),
        migrations.RemoveField(
            model_name='buglink',
            name='test_result',
        ),
        migrations.RemoveField(
            model_name='buglink',
            name='test_runs',
        ),
        migrations.RemoveField(
            model_name='bundle',
            name='bundle_stream',
        ),
        migrations.RemoveField(
            model_name='bundle',
            name='uploaded_by',
        ),
        migrations.RemoveField(
            model_name='bundlestream',
            name='group',
        ),
        migrations.RemoveField(
            model_name='bundlestream',
            name='user',
        ),
        migrations.RemoveField(
            model_name='image',
            name='filter',
        ),
        migrations.RemoveField(
            model_name='imagechartfilter',
            name='filter',
        ),
        migrations.RemoveField(
            model_name='imagechartfilter',
            name='image_chart',
        ),
        migrations.AlterUniqueTogether(
            name='imagecharttest',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='imagecharttest',
            name='image_chart_filter',
        ),
        migrations.RemoveField(
            model_name='imagecharttest',
            name='test',
        ),
        migrations.RemoveField(
            model_name='imagecharttestattribute',
            name='image_chart_test',
        ),
        migrations.AlterUniqueTogether(
            name='imagecharttestattributeuser',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='imagecharttestattributeuser',
            name='image_chart_test_attribute',
        ),
        migrations.RemoveField(
            model_name='imagecharttestattributeuser',
            name='user',
        ),
        migrations.AlterUniqueTogether(
            name='imagecharttestcase',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='imagecharttestcase',
            name='image_chart_filter',
        ),
        migrations.RemoveField(
            model_name='imagecharttestcase',
            name='test_case',
        ),
        migrations.RemoveField(
            model_name='imagecharttestcaseattribute',
            name='image_chart_test_case',
        ),
        migrations.AlterUniqueTogether(
            name='imagecharttestcaseuser',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='imagecharttestcaseuser',
            name='image_chart_test_case',
        ),
        migrations.RemoveField(
            model_name='imagecharttestcaseuser',
            name='user',
        ),
        migrations.AlterUniqueTogether(
            name='imagecharttestuser',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='imagecharttestuser',
            name='image_chart_test',
        ),
        migrations.RemoveField(
            model_name='imagecharttestuser',
            name='user',
        ),
        migrations.AlterUniqueTogether(
            name='imagechartuser',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='imagechartuser',
            name='image_chart',
        ),
        migrations.RemoveField(
            model_name='imagechartuser',
            name='user',
        ),
        migrations.RemoveField(
            model_name='imagereport',
            name='group',
        ),
        migrations.RemoveField(
            model_name='imagereport',
            name='image_report_group',
        ),
        migrations.RemoveField(
            model_name='imagereport',
            name='user',
        ),
        migrations.AlterUniqueTogether(
            name='imagereportchart',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='imagereportchart',
            name='image_report',
        ),
        migrations.RemoveField(
            model_name='imageset',
            name='images',
        ),
        migrations.AlterUniqueTogether(
            name='namedattribute',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='namedattribute',
            name='content_type',
        ),
        migrations.RemoveField(
            model_name='pmqabundlestream',
            name='bundle_stream',
        ),
        migrations.DeleteModel(
            name='SoftwarePackageScratch',
        ),
        migrations.AlterUniqueTogether(
            name='testcase',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='testcase',
            name='test',
        ),
        migrations.DeleteModel(
            name='TestDefinition',
        ),
        migrations.RemoveField(
            model_name='testresult',
            name='test_case',
        ),
        migrations.RemoveField(
            model_name='testresult',
            name='test_run',
        ),
        migrations.RemoveField(
            model_name='testrun',
            name='bundle',
        ),
        migrations.RemoveField(
            model_name='testrun',
            name='devices',
        ),
        migrations.RemoveField(
            model_name='testrun',
            name='packages',
        ),
        migrations.RemoveField(
            model_name='testrun',
            name='sources',
        ),
        migrations.RemoveField(
            model_name='testrun',
            name='tags',
        ),
        migrations.RemoveField(
            model_name='testrun',
            name='test',
        ),
        migrations.DeleteModel(
            name='SoftwarePackage',
        ),
        migrations.AlterUniqueTogether(
            name='testrunfilter',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='testrunfilter',
            name='bundle_streams',
        ),
        migrations.RemoveField(
            model_name='testrunfilter',
            name='owner',
        ),
        migrations.RemoveField(
            model_name='testrunfilter',
            name='uploaded_by',
        ),
        migrations.RemoveField(
            model_name='testrunfilterattribute',
            name='filter',
        ),
        migrations.AlterUniqueTogether(
            name='testrunfiltersubscription',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='testrunfiltersubscription',
            name='filter',
        ),
        migrations.RemoveField(
            model_name='testrunfiltersubscription',
            name='user',
        ),
        migrations.RemoveField(
            model_name='testrunfiltertest',
            name='filter',
        ),
        migrations.RemoveField(
            model_name='testrunfiltertest',
            name='test',
        ),
        migrations.RemoveField(
            model_name='testrunfiltertestcase',
            name='test',
        ),
        migrations.RemoveField(
            model_name='testrunfiltertestcase',
            name='test_case',
        ),
        migrations.DeleteModel(
            name='Attachment',
        ),
        migrations.DeleteModel(
            name='BugLink',
        ),
        migrations.DeleteModel(
            name='Bundle',
        ),
        migrations.DeleteModel(
            name='BundleDeserializationError',
        ),
        migrations.DeleteModel(
            name='BundleStream',
        ),
        migrations.DeleteModel(
            name='HardwareDevice',
        ),
        migrations.DeleteModel(
            name='Image',
        ),
        migrations.DeleteModel(
            name='ImageChartFilter',
        ),
        migrations.DeleteModel(
            name='ImageChartTest',
        ),
        migrations.DeleteModel(
            name='ImageChartTestAttribute',
        ),
        migrations.DeleteModel(
            name='ImageChartTestAttributeUser',
        ),
        migrations.DeleteModel(
            name='ImageChartTestCase',
        ),
        migrations.DeleteModel(
            name='ImageChartTestCaseAttribute',
        ),
        migrations.DeleteModel(
            name='ImageChartTestCaseUser',
        ),
        migrations.DeleteModel(
            name='ImageChartTestUser',
        ),
        migrations.DeleteModel(
            name='ImageChartUser',
        ),
        migrations.DeleteModel(
            name='ImageReport',
        ),
        migrations.DeleteModel(
            name='ImageReportChart',
        ),
        migrations.DeleteModel(
            name='ImageReportGroup',
        ),
        migrations.DeleteModel(
            name='ImageSet',
        ),
        migrations.DeleteModel(
            name='NamedAttribute',
        ),
        migrations.DeleteModel(
            name='PMQABundleStream',
        ),
        migrations.DeleteModel(
            name='SoftwareSource',
        ),
        migrations.DeleteModel(
            name='Tag',
        ),
        migrations.DeleteModel(
            name='Test',
        ),
        migrations.DeleteModel(
            name='TestCase',
        ),
        migrations.DeleteModel(
            name='TestResult',
        ),
        migrations.DeleteModel(
            name='TestRun',
        ),
        migrations.DeleteModel(
            name='TestRunDenormalization',
        ),
        migrations.DeleteModel(
            name='TestRunFilter',
        ),
        migrations.DeleteModel(
            name='TestRunFilterAttribute',
        ),
        migrations.DeleteModel(
            name='TestRunFilterSubscription',
        ),
        migrations.DeleteModel(
            name='TestRunFilterTest',
        ),
        migrations.DeleteModel(
            name='TestRunFilterTestCase',
        ),
    ]
